from pathlib import Path
from typing import Optional

from pydantic import AnyUrl, BaseModel, DirectoryPath, FileUrl, PostgresDsn
from sqlmodel import Field, SQLModel


class SshUrl(AnyUrl):
    allowed_schemes = {"ssh"}
    user_required = True


class SshRemote(BaseModel):
    url: SshUrl


class SqliteDsn(FileUrl):
    allowed_schemes = {"sqlite"}


class SqliteConn(BaseModel):
    dir: DirectoryPath
    file: Path
    url: SqliteDsn

    @classmethod
    def builder(cls, dir: DirectoryPath, file: Path):
        url: SqliteDsn = f"sqlite:///{dir}/{file}"  # pyright:ignore
        return cls(dir=dir, file=file, url=url)


class PostgresConn(BaseModel):
    url: PostgresDsn


class Host(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)



import uvicorn
from .v1.app import app as api


def uvicorn_local(**kwargs):
    uvicorn.run(api, **kwargs)


def gunicorn_full():
    print("Starting Gunicorn... NOT")
    pass


def main():
    gunicorn_full()



from typing import Optional
from pydantic import BaseSettings, PostgresDsn
from pathlib import Path

DEFAULT_DATA_DIR = Path.home() / ".shipyard"
DEFAULT_SQLITE_FILENAME = Path("shipyard.db")


class Settings(BaseSettings):
    data_dir: Path = DEFAULT_DATA_DIR
    sqlite_filename: Path = DEFAULT_SQLITE_FILENAME
    postgres_url: Optional[PostgresDsn] = None

    class Config:
        env_prefix = "shipyard_"
        env_file = ".env"


_settings: Optional[Settings] = None


def load_settings(**kwargs) -> Settings:
    global _settings
    if _settings:
        raise Exception("Settings must be loaded only once.")
    else:
        _settings = Settings(**kwargs)
    return _settings


def get_settings() -> Settings:
    global _settings
    if not _settings:
        raise Exception("Settings must be loaded before being used.")
    return _settings



import os
from pathlib import Path
from rich import print
import typer

from shipyard.api.main import gunicorn_full, uvicorn_local
from shipyard.db import connect_database
from shipyard.envoy.tunnel import open_tunnel
from shipyard.models import SshRemote
from shipyard.settings import get_settings, load_settings, DEFAULT_DATA_DIR


app = typer.Typer()


@app.callback(no_args_is_help=True)
def callback(
    data_dir: str = typer.Option(
        None,
        help="Directory containing the application database, will be created if necessary.",
        show_default=f"{DEFAULT_DATA_DIR}{os.sep}",  # pyright:ignore
    ),
):
    """
    Urbit hosting and automation platform

    Global options must be entered before a command.
    """
    if data_dir:
        load_settings(data_dir=data_dir)
    else:
        load_settings()

    ensure_data_dir()

    connect_database()


def ensure_data_dir():
    data_dir: Path = get_settings().data_dir

    if data_dir.is_file():
        print(f"Cannot access directory {data_dir}{os.sep}: It is a regular file.")
        raise typer.Exit(1)

    if data_dir.is_symlink() and not data_dir.exists():
        print(f"Cannot access directory {data_dir}{os.sep}: It is a broken link.")
        raise typer.Exit(1)

    if not data_dir.is_dir():
        proceed = typer.confirm(
            f"Create data directory?: {data_dir}{os.sep}",
            default=(data_dir == DEFAULT_DATA_DIR),
            abort=True,
        )
        if proceed:
            try:
                data_dir.mkdir()
            except Exception as e:
                print(f"Cannot create directory {data_dir}{os.sep}: {e}.")
                raise typer.Exit(1)


@app.command()
def api(
    host: str = typer.Option("127.0.0.1", help="Host the server will listen on"),
    port: int = typer.Option("8000", help="Port the server will listen on"),
    log_level: str = typer.Option("info", help="Logging level of the server"),
    full: bool = typer.Option(False, help="Run a full multi-worker Gunicorn server"),
):
    """
    Run a local API server
    """
    if full:
        gunicorn_full()
    else:
        uvicorn_local(host=host, port=port, log_level=log_level)


@app.command(no_args_is_help=True)
def tunnel(
    host: str = typer.Option(
        ..., "--host", "-h", help="Hostname of the remote SSH server"
    ),
    user: str = typer.Option(
        ..., "--user", "-u", help="Username to login to remote SSH server"
    ),
    port: int = typer.Option(
        22, "--port", "-p", help="Listening port of the remote SSH server"
    ),
    local_port: int = typer.Option(
        4000, "--local-port", "-L", help="Local port to forward application to"
    ),
    remote_port: int = typer.Option(
        4000, "--remote-port", "-R", help="Remote application port to forward"
    ),
    verbose: bool = typer.Option(
        False, "--verbose", "-v", help="Print additional status information"
    ),
):
    """
    Open an SSH tunnel to a remote ship, allowing you to connect on a local port
    """
    ssh: SshRemote = SshRemote(url=f"ssh://{user}@{host}:{port}")  # pyright:ignore
    open_tunnel(ssh, local_port, remote_port, verbose)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['shipyard',
 'shipyard.api',
 'shipyard.api.v1',
 'shipyard.cli',
 'shipyard.colony',
 'shipyard.deploy',
 'shipyard.envoy',
 'shipyard.vigil']

package_data = \
{'': ['*']}

install_requires = \
['docker[ssh]>=5.0.3,<6.0.0',
 'fastapi>=0.79.0,<0.80.0',
 'psycopg2-binary>=2.9.3,<3.0.0',
 'python-dotenv>=0.21.0,<0.22.0',
 'sqlmodel>=0.0.8,<0.0.9',
 'typer[all]>=0.6.1,<0.7.0',
 'uvicorn[standard]>=0.18.2,<0.19.0']

entry_points = \
{'console_scripts': ['shipyard = shipyard.cli.main:app']}

setup_kwargs = {
    'name': 'shipyard-urbit',
    'version': '0.1.2',
    'description': 'Urbit hosting and automation platform',
    'long_description': '# Shipyard\n\nUrbit hosting and automation platform.\n\nNote: this is a Pre-Release package.  All changes will be breaking.  Wait for release 1.0.0 or later.\n\n## Install\n\n```\npip install shipyard-urbit\n```\n\n## Usage\n\n```\nshipyard\n```\n\n## Configuration\n\nSpecify the following environment vars to configure your application.  You may use a `.env` file in the location where `shipyard` is run.\n\n- `SHIPYARD_DATA_DIR` - directory where SQLite database and other data will live, default: `~/.shipyard/`\n    -  Override at runtime using the `--data-dir` option\n- `SHIPYARD_SQLITE_FILENAME` - name of the db file within the data directory, default: `shipyard.db`\n- `SHIPYARD_POSTGRES_URL` - PostgreSQL connection string to override use of SQLite, default: `None`\n\nPostgreSQL is only recommended for large multi-user deployments. You must create your database before connecting with shipyard like so:\n\n```\nCREATE DATABASE shipyard;\n```\n\nThen `SHIPYARD_POSTGRES_URL` should look something like this:\n\n```\npostgresql://user:password@127.0.0.1:5432/shipyard\n```\n\n## API Overview\n\nVisit [redacted] for full API Documentation.\n\n## Development\n\n### Modules\n\n#### shipyard\n\n * `db.py` - database engine for sqlite/postresql application state\n * `models.py` - types used throughout the project\n * `settings.py` - system-wide settings and env vars\n\n#### shipyard.api\n\nHTTP API built with FastAPI.\n\n#### shipyard.cli\n\nCommand-line interface built with Typer.\n\n#### shipyard.colony\n\nHost setup and configuration using Ansible.\n\n#### shipyard.deploy\n\nCreating and migrating Urbit ships within our host infrastructure.\n\n#### shipyard.envoy\n\nCommunication and direction of Urbit ships.\n\n#### shipyard.vigil\n\nMonitoring and alerting. WIP.\n\n## License\n\nThis project is licensed under Apache-2.0.  Code licensed from other projects will be clearly marked with the appropriate notice.\n',
    'author': 'nodreb-borrus',
    'author_email': 'nodreb@borr.us',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Optional, Union

from sqlmodel import SQLModel, create_engine
from sqlalchemy.future import Engine

from .models import PostgresConn, SqliteConn
from .settings import get_settings


_engine: Optional[Engine] = None


def connect_database():
    global _engine
    if _engine:
        raise Exception("Database engine should be created only once.")

    conn: Union[PostgresConn, SqliteConn]
    settings = get_settings()
    if settings.postgres_url:
        conn = PostgresConn(url=settings.postgres_url)
    else:
        conn = SqliteConn.builder(settings.data_dir, settings.sqlite_filename)

    _engine = create_engine(conn.url)

    SQLModel.metadata.create_all(_engine)



#!/usr/bin/env python

# Copyright (C) 2003-2007  Robey Pointer <robeypointer@gmail.com>
#
# This file is part of paramiko.
#
# Paramiko is free software; you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# Paramiko is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Paramiko; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA.


import sys
import select
import socketserver as SocketServer
import paramiko
from shipyard.models import SshRemote

SSH_PORT = 22
DEFAULT_PORT = 4000

g_verbose = True


class ForwardServer(SocketServer.ThreadingTCPServer):
    daemon_threads = True
    allow_reuse_address = True


class Handler(SocketServer.BaseRequestHandler):
    def handle(self):
        try:
            chan = self.ssh_transport.open_channel(
                "direct-tcpip",
                (self.chain_host, self.chain_port),
                self.request.getpeername(),
            )
        except Exception as e:
            verbose(
                "Incoming request to %s:%d failed: %s"
                % (self.chain_host, self.chain_port, repr(e))
            )
            return
        if chan is None:
            verbose(
                "Incoming request to %s:%d was rejected by the SSH server."
                % (self.chain_host, self.chain_port)
            )
            return

        verbose(
            "Connected!  Tunnel open %r -> %r -> %r"
            % (
                self.request.getpeername(),
                chan.getpeername(),
                (self.chain_host, self.chain_port),
            )
        )
        while True:
            r, w, x = select.select([self.request, chan], [], [])
            if self.request in r:
                data = self.request.recv(1024)
                if len(data) == 0:
                    break
                chan.send(data)
            if chan in r:
                data = chan.recv(1024)
                if len(data) == 0:
                    break
                self.request.send(data)

        peername = self.request.getpeername()
        chan.close()
        self.request.close()
        verbose("Tunnel closed from %r" % (peername,))


def forward_tunnel(local_port, remote_host, remote_port, transport):
    # this is a little convoluted, but lets me configure things for the Handler
    # object.  (SocketServer doesn't give Handlers any way to access the outer
    # server normally.)
    class SubHander(Handler):
        chain_host = remote_host
        chain_port = remote_port
        ssh_transport = transport

    ForwardServer(("", local_port), SubHander).serve_forever()


def verbose(s):
    if g_verbose:
        print(s)


def open_tunnel(
    ssh: SshRemote,
    local_port: int = DEFAULT_PORT,
    remote_port: int = DEFAULT_PORT,
    verb: bool = False,
):
    host = ssh.url.host if ssh.url.host else ""
    port = int(ssh.url.port) if ssh.url.port else SSH_PORT
    user = ssh.url.user if ssh.url.user else ""
    global g_verbose
    g_verbose = verb

    client = paramiko.SSHClient()
    client.load_system_host_keys()
    client.set_missing_host_key_policy(paramiko.WarningPolicy())

    print(f"Connecting to {host}:{port} ...")

    try:
        client.connect(
            host,
            port,
            username=user,
            look_for_keys=True,
            password=None,
        )
    except Exception as e:
        print("*** Failed to connect to %s:%s: %r" % (host, port, e))
        sys.exit(1)

    print("Forwarding local port %d to %s:%d ..." % (local_port, host, remote_port))

    try:
        forward_tunnel(local_port, "localhost", remote_port, client.get_transport())
    except KeyboardInterrupt:
        print(" Port forwarding stopped.")
        sys.exit(0)
